'''
CommonPy: an assortment of Python helper functions and utility classes

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

# Package metadata ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
#  ╭────────────────────── Notice ── Notice ── Notice ─────────────────────╮
#  |    The following values are automatically updated at every release    |
#  |    by the Makefile. Manual changes to these values will be lost.      |
#  ╰────────────────────── Notice ── Notice ── Notice ─────────────────────╯

__version__     = '1.3.1'
__description__ = 'Assortment of Python helper functions and utility classes'
__url__         = 'https://github.com/caltechlibrary/commonpy'
__author__      = 'Michael Hucka'
__email__       = 'mhucka@caltech.edu'
__license__     = 'BSD 3-clause'


# Miscellaneous utilities.
# .............................................................................

def print_version():
    print(f'{__name__} version {__version__}')
    print(f'Authors: {__author__}')
    print(f'URL: {__url__}')
    print(f'License: {__license__}')



'''
module_utils: utilities related to Python modules and packages

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import os
from os import path as path
import sys


# Constants.
# .............................................................................

_APP_REG_PATH_FORMAT = r'Software\Caltech Library\{}\Settings'
'''Format of the the Windows registry path for this application.  Contains one
string format substitution placeholder.  Use this with a call to format(...)
like this: _APP_REG_PATH_FORMAT.format(module).
'''


# Main functions.
# .............................................................................

def module_path(module_name = __package__):
    '''Returns the absolute path to the installation directory of the Python
    module named 'name'.  The name defaults to __package__.
    '''
    if module_name in sys.modules:
        return path.abspath(sys.modules[module_name].__path__[0])
    else:
        return None


def datadir_path(module_name = __package__):
    '''Returns the path to the named module's internal data directory.'''
    return path.join(module_path(module_name), 'data')


def desktop_path():
    '''Returns the path to the user's desktop directory.'''
    if sys.platform.startswith('win'):
        return path.join(os.environ['USERPROFILE'], 'Desktop')
    else:
        return path.join(path.expanduser('~'), 'Desktop')


def installation_path(module_name = __package__):
    '''Returns the path to where the given Python package is installed.'''
    # The path returned by module.__path__ is to the directory containing
    # the __init__.py file.  What we want here is the path to the installation
    # of the application binary.
    if sys.platform.startswith('win'):
        from winreg import OpenKey, CloseKey, QueryValueEx, HKEY_LOCAL_MACHINE, KEY_READ
        try:
            if __debug__: log('reading Windows registry entry')
            reg_path = _APP_REG_PATH_FORMAT.format(module_name)
            key = OpenKey(HKEY_LOCAL_MACHINE, reg_path)
            value, regtype = QueryValueEx(key, 'Path')
            CloseKey(key)
            if __debug__: log(f'path to windows installation: {value}')
            return value
        except WindowsError:
            # Kind of a problem. Punt and return a default value.
            default_path = path.abspath('C:\\Program Files\\{}'.format(module_name))
            if __debug__: log(f'defaulting to {default_path}')
            return default_path
    else:
        return path.abspath(path.join(module_path(module_name), '..'))


def config_path(module_name = __package__):
    '''Returns the path to ~/.config or equivalent.'''
    if sys.platform.startswith('win'):
        config_root = path.join(os.environ['USERPROFILE'], 'AppData', 'Local')
    else:
        config_root = path.join(path.expanduser('~'), '.config')
    return path.join(config_root, module_name)



'''
network_utils.py: network utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2018-2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import httpx
import socket
import ssl
import tldextract
import urllib

if __debug__:
    from sidetrack import log

from .interrupt import wait, interrupted


# Internal constants.
# .............................................................................

_MAX_RECURSIVE_CALLS = 10
'''How many times can certain network functions call themselves upon
encountering a network error before they stop and give up.'''

_MAX_CONSECUTIVE_FAILS = 3
'''Maximum number of consecutive failures before pause and try another round.'''

_MAX_RETRIES = 5
'''Maximum number of times we back off and try again.  This also affects the
maximum wait time that will be reached after repeated retries.'''


# Main functions.
# .............................................................................

def network_available(address = "8.8.8.8", port = 53, timeout = 5):
    '''Return True if it appears we have a network connection, False if not.
    By default, this attempts to contact one of the Google DNS servers (as a
    plain TCP connection, not as an actual DNS lookup).  Argument 'address'
    and 'port' can be used to test a different server address and port.  The
    socket connection is attempted for 'timeout' seconds.
    '''
    # Portions of this code are based on the answer by user "7h3rAm" posted to
    # Stack Overflow here: https://stackoverflow.com/a/33117579/743730
    try:
        if __debug__: log('testing if we have a network connection')
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((address, port))
        if __debug__: log('we have a network connection')
        return True
    except Exception:
        if __debug__: log('could not connect to https://www.google.com')
        return False


def hostname(url):
    if url.startswith('http'):
        parsed = urllib.parse.urlsplit(url)
        return parsed.hostname
    else:
        # urllib.parse doesn't provide a hostname.  Try a different way.
        return '.'.join(part for part in tldextract.extract(url) if part)


def scheme(url):
    parsed = urllib.parse.urlsplit(url)
    return parsed.scheme


def netloc(url):
    parsed = urllib.parse.urlsplit(url)
    if parsed.netloc:
        return parsed.netloc
    elif not url.startswith('http') and '//' not in url:
        # Add a fake "http://" so that urllib.parse can figure out the netloc.
        parsed = urllib.parse.urlsplit('http://' + url)
        return parsed.netloc
    else:
        # Last-ditch effort.
        return parsed.path


def timed_request(method, url, client = None, **kwargs):
    '''Perform a network access, automatically retrying if exceptions occur.

    The value given to parameter "method" must be a string chosen from among
    valid HTTP methods, such as "get", "post", or "head".  If "client" is
    not None, it is used as an httpx.Client object. Other keyword arguments
    are passed to the network method.

    If not given a Client object, the default timeouts for network connect,
    read, and write are 15 seconds.  It enables HTTP 2.0 and disables SSL
    verification.

    This method retries connections in cases of network exceptions.  It also
    retries connections one time when the server returns certain HTTP status
    codes, specifically 400, 409, 502, 503, and 504.  These are sometimes the
    result of temporary server problems or other issues and disappear when a
    second attempt is made after a brief pause.
    '''
    def addurl(text):
        return f'{text} for {url}'

    if client is None:
        timeout = httpx.Timeout(15, connect = 15, read = 15, write = 15)
        client = httpx.Client(timeout = timeout, http2 = True, verify = False)

    failures = 0
    retries = 0
    error = None
    while failures < _MAX_CONSECUTIVE_FAILS and not interrupted():
        try:
            if __debug__: log(addurl(f'doing http {method}'))
            func = getattr(client, method)
            response = func(url, **kwargs)
            # For some statuses, retry once, in case it's a transient problem.
            code = response.status_code
            if __debug__: log(addurl(f'got response with code {code}'))
            if code not in [400, 409, 502, 503, 504] or failures > 0:
                return response
            else:
                failures += 1
        except KeyboardInterrupt as ex:
            if __debug__: log(addurl(f'network {method} interrupted by {str(ex)}'))
            raise
        except (httpx.CookieConflict, httpx.StreamError, httpx.TooManyRedirects,
                httpx.DecodingError, httpx.ProtocolError, httpx.ProxyError) as ex:
            # Probably indicates a deeper issue.  Don't do our lengthy retry
            # sequence, but try one more time, in case it's transient.
            if __debug__: log(addurl(f'exception {str(ex)}'))
            if failures > 0:
                raise
            failures += 1
            if __debug__: log(addurl('retrying one more time after brief pause'))
        except Exception as ex:
            # Problem might be transient.  Don't quit right away.
            failures += 1
            if __debug__: log(addurl(f'exception (failure #{failures}): {str(ex)}'))
            # Record the first error we get, not the subsequent ones, because
            # in the case of network outages, the subsequent ones will be
            # about being unable to reconnect and not the original problem.
            if not error:
                error = ex
        if failures >= _MAX_CONSECUTIVE_FAILS:
            # Pause with exponential back-off, reset failure count & try again.
            if retries < _MAX_RETRIES:
                retries += 1
                failures = 0
                pause = 10 * retries * retries
                if __debug__: log(addurl(f'pausing due to consecutive failures'))
                wait(pause)
            else:
                if __debug__: log(addurl('exceeded max failures and max retries'))
                raise error
        # Pause briefly b/c it's rarely a good idea to retry immediately.
        if __debug__: log(addurl('pausing briefly before retrying'))
        wait(0.5)
    if interrupted():
        if __debug__: log(addurl('interrupted'))
        raise Interrupted(addurl('Network request has been interrupted'))
    else:
        # In theory, we should never reach this point.  If we do, then:
        raise InternalError(addurl('Unexpected case in timed_request'))


def net(method, url, client = None, handle_rate = True,
        polling = False, recursing = 0, **kwargs):
    '''Invoke HTTP "method" on 'url' with optional keyword arguments provided.

    Returns a tuple of (response, exception), where the first element is
    the response from the get or post http call, and the second element is
    an exception object if an exception occurred.  If no exception occurred,
    the second element will be None.  This allows the caller to inspect the
    response even in cases where exceptions are raised.

    If keyword 'client' is not None, it's assumed to be a Python HTTPX Client
    object to use for the network call.  Settings such as timeouts should be
    done by the caller creating appropriately-configured Client objects.

    If keyword 'handle_rate' is True, this function will automatically pause
    and retry if it receives an HTTP code 429 ("too many requests") from the
    server.  If False, it will return the exception RateLimitExceeded instead.

    If keyword 'polling' is True, certain statuses like 404 are ignored and
    the response is returned; otherwise, they are considered errors.  The
    behavior when True is useful in situations where a URL does not exist until
    something is ready at the server, and the caller is repeatedly checking
    the URL.  It is up to the caller to implement the polling schedule and
    call this function (with polling = True) as needed.

    This method always passes the argument allow_redirects = True to the
    underlying Python HTTPX library network calls.
    '''
    def addurl(text):
        return f'{text} for {url}'

    resp = None
    try:
        resp = timed_request(method, url, client, allow_redirects = True, **kwargs)
    except (httpx.NetworkError, httpx.ProtocolError) as ex:
        # timed_request() will have retried, so if we get here, time to bail.
        if __debug__: log(addurl(f'got network exception: {str(ex)}'))
        if not network_available():
            if __debug__: log(addurl('returning NetworkFailure'))
            return (resp, NetworkFailure(addurl('Network connectivity failure')))
        else:
            if __debug__: log(addurl('failed > 1 times -- returning ServiceFailure'))
            return (resp, ServiceFailure(addurl('Network or server error {str(ex)}')))
    except Exception as ex:
        # Not a network or protocol error, and not a normal server response.
        if __debug__: log(addurl(f'returning exception: {str(ex)}'))
        return (resp, ex)

    # Interpret the response.  Note that the requests library handles code 301
    # and 302 redirects automatically, so we don't need to do it here.
    error = None
    code = resp.status_code
    reason = resp.reason_phrase
    if code == 400:
        error = ServiceFailure(addurl('Server rejected the request'))
    elif code in [401, 402, 403, 407, 451, 511]:
        error = AuthenticationFailure(addurl('Access is forbidden'))
    elif code in [404, 410] and not polling:
        error = NoContent(addurl("No content found"))
    elif code in [405, 406, 409, 411, 412, 413, 414, 417, 428, 431, 505, 510]:
        error = InternalError(addurl(f'Server returned code {code} ({reason})'))
    elif code in [415, 416]:
        error = ServiceFailure(addurl('Server rejected the request ({reason})'))
    elif code == 429:
        if handle_rate and recursing < _MAX_RECURSIVE_CALLS:
            pause = 5 * (recursing + 1)   # +1 b/c we start with recursing = 0.
            if __debug__: log(addurl('rate limit hit -- pausing'))
            wait(pause)                   # 5 s, then 10 s, then 15 s, etc.
            if __debug__: log(addurl(f'doing recursive call #{recursing + 1}'))
            return net(method, url, client, handle_rate, polling, recursing + 1, **kwargs)
        error = RateLimitExceeded(addurl('Server blocking requests due to rate limits'))
    elif code in [500, 501, 502, 503, 504, 506, 507, 508]:
        error = ServiceFailure(addurl(f'Server error (code {code} -- {reason})'))
    elif not (200 <= code < 400):
        error = NetworkFailure(addurl(f'Unable to resolve {url}'))
    # The error msg will have had the URL added already; no need to do it here.
    if __debug__: log('returning {}'.format(f'{error}' if error else f'response for {url}'))
    return (resp, error)



'''
string_utils: string manipulation utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


# Functions.
# .............................................................................

def antiformat(s):
    '''Quote instances of '{' and '}' in 's' so it can be passed to format.'''

    # This is a combo of 2 solutions posted by Stack Overflow users:
    # Kristján Valur (2019-12-17): https://stackoverflow.com/a/59371315/743730
    # "clickbait" (2018-07-31): https://stackoverflow.com/a/51623363/743730

    s = str(s)
    s = s.replace('{{', '⁌').replace('}}', '⁍')
    s = s.replace('{', '{{').replace('}', '}}')
    return s.replace('⁌', '{').replace('⁍', '}')



'''
data_utils: operating system & platform utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   boltons import ecoutils
import yaml


# Functions.
# .............................................................................

def system_profile(anonymize = False, as_dict = False):
    '''Return a summary of this system.

    The default output format is a string that is more or less human readable.
    If the value of parameter "as_dict" is True, the output is a Python dict.

    If the value of parameter "anonymize" is True, the results will have values
    of '-' for certain identifiable information.  This includes current working
    directory, hostname, Python executable path, command-line arguments, and
    username.

    The output of this function is identical to that of "get_profile" from the
    Boltons ecoutils module, reformatted and returned as a string that can be
    printed for human consumption.
    '''

    details = ecoutils.get_profile(scrub = anonymize)
    return details if as_dict else yaml.dump(details)



'''
data_utils: data manipulation utilities

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   boltons.strutils import pluralize
import dateparser
import datetime
from   datetime import datetime as dt
from   dateutil import tz
from   humanize import intcomma


# Constants.
# .............................................................................

DATE_FORMAT = '%b %d %Y %H:%M:%S %Z'
'''Format in which lastmod date is printed back to the user. The value is used
with datetime.strftime().'''


# Functions.
# .............................................................................

def slice(lst, n):
    '''Yield n number of slices from lst.'''
    # Original algorithm from Jurgen Strydom posted 2019-02-21 Stack Overflow
    # https://stackoverflow.com/a/54802737/743730
    for i in range(0, n):
        yield lst[i::n]


def unique(lst):
    '''Take list "lst" and return a version without duplicates.'''
    # This exists because I think the intention behind list(set(...)) is
    # too easily lost and makes code muddier.
    return list(set(lst))


def ordinal(n):
    '''Print a number followed by "st" or "nd" or "rd", as appropriate.'''
    # Spectacular algorithm by user "Gareth" at this posting:
    # http://codegolf.stackexchange.com/a/4712
    return '{}{}'.format(n, 'tsnrhtdd'[(n/10%10!=1)*(n%10<4)*n%10::4])


def expanded_range(text):
    '''Return individual numbers for a range expressed as X-Y.'''
    # This makes the range 1-100 be 1, 2, ..., 100 instead of 1, 2, ..., 99
    if '-' in text:
        range_list = text.split('-')
        # Malformed cases of -x, where first number is missing.  Take it as 1.
        if not range_list[0].isdigit():
            range_list = [1, range_list[1]]
        # Malformed cases of x-, where 2nd number missing.  Can't handle this.
        if not range_list[1].isdigit():
            raise ValueError(f'Malformed range expression: "{text}"')
        range_list.sort(key = int)
        return [*map(str, range(int(range_list[0]), int(range_list[1]) + 1))]
    else:
        return text


def timestamp():
    '''Return a string describing the date and time right now.'''
    return dt.now(tz = tz.tzlocal()).strftime(DATE_FORMAT)


def parsed_datetime(string):
    '''Parse a human-written time/date string using dateparser's parse()
function with predefined settings.'''
    return dateparser.parse(string, settings = {'RETURN_AS_TIMEZONE_AWARE': True})


def pluralized(word, count, include_number = False):
    '''Pluralize the "word" if "count" is > 1 or has length > 1.

    If "include_number" is true, then the value of "count" will be prepended
    before the string:

        files_list = ['one.txt', 'two.txt', 'three.txt']
        plural('file', files_list)       --> 'files'
        plural('file', files_list, True) --> '3 files'

    This function is useful in f-strings when words refer to a number of items
    whose total or length is only known at run time.  For example,

       f"Uploading {pluralized('file', files_list, True)} to server."
    '''

    if isinstance(count, int):
        num = count
    elif isinstance(count, (list, set, dict)) or type(count) is {}.values().__class__:
        num = len(count)
    else:
        # If we can't figure out what kind of thing we're counting, we can't
        # pluralize, so just return the word as-is.
        return word
    text = pluralize(word) if num > 1 else word
    if include_number:
        return f'{intcomma(num)} {text}'
    else:
        return text



'''
exceptions.py: exceptions defined by CommonPy

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


# Base class.
# .............................................................................
# The base class makes it possible to use a single test to distinguish
# between exceptions generated by CommonPy code and exceptions generated by
# something else.

class CommonPyException(Exception):
    '''Base class for CommonPy exceptions.'''
    pass


# Exception classes.
# .............................................................................

class Interrupted(CommonPyException):
    '''The user elected to cancel/quit the program.'''
    pass

class NetworkFailure(CommonPyException):
    '''Unrecoverable problem involving network operations.'''
    pass

class ServiceFailure(CommonPyException):
    '''Unrecoverable problem involving a remote service.'''
    pass

class AuthenticationFailure(CommonPyException):
    '''Problem obtaining or using authentication credentials.'''
    pass

class NoContent(CommonPyException):
    '''No content found at the given location.'''
    pass

class RateLimitExceeded(CommonPyException):
    '''The service flagged reports that its rate limits have been exceeded.'''
    pass

class InternalError(CommonPyException):
    '''Unrecoverable problem involving commonpy itself.'''
    pass



'''
interrupt.py: provide an interruptible wait(...) and related utilities.

This module includes wait(...), a replacement for sleep(...) that is
interruptible and works with multiple threads.  It also provides methods to
cause an interruption (including doing it by issuing a ^C to the program),
check whether an interruption occurred, and other related operations.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import signal
import sys
import threading

if sys.platform == "win32":
    import win32api

if __debug__:
    from sidetrack import log


# Global variables.
# .............................................................................

__waiter = threading.Event()
'''Internal state variable used to communicate that an interrupt occurred.'''

__exception = KeyboardInterrupt
'''Exception that will be raised when an interrupt signal is received.'''


# Main functions.
# .............................................................................

def config_interrupt(callback = None, raise_exception = KeyboardInterrupt,
                     on_signal = signal.SIGINT):
    '''Configure handling of interruptions.

    This function can be used to attach a signal handler that will do the
    following upon receipt of a designated signal:

      1. call interrupt(), defined elsewhere in this module
      2. optionally call the function "callback" given as argument
      3. raise the given "exception" (default: KeyboardInterrupt)

    Parameter "callback" assigns a function to be called after calling
    interrupt() when the signal is received.  If no value is given for
    "callback", no callback function is called.

    Parameter "raise_exception" determines the exception that is raised after
    calling the callback function.

    Parameter "signal" determines the signal to which the handler is bound.
    By default, signal.SIGINT is used, which is ^C on Unix-like systems.
    Note that changing this to something *other* than SIGINT means that ^C
    will then trigger the normal Python keyboard interrupt signal handling
    instead of invoking this signal handler.
    '''

    global __exception
    __exception = raise_exception

    def interrupt_handler(signum, frame):
        if __debug__: log('interrupt_handler invoked')
        interrupt()
        if callback:
            if __debug__: log(f'invoking callback {callback}')
            callback()
        if __debug__: log(f'raising {raise_exception}')
        raise __exception

    if __debug__: log(f'setting handler on {on_signal}, exception {__exception}')
    if sys.platform == "win32":
        win32api.SetConsoleCtrlHandler(interrupt_handler, True)
    else:
        signal.signal(on_signal, interrupt_handler)


def wait(duration):
    '''Wait for "duration" seconds, in a way that can be interrupted.

    This is a replacement for sleep(duration).  If interrupted, this function
    raises the exception configured by a prior call to config_interrupt(...).
    '''
    if __debug__: log(f'waiting for {duration} s')
    __waiter.wait(duration)
    if interrupted():
        if __debug__: log(f'raising {__exception}')
        raise __exception


def interrupt():
    '''Interrupt any waits and internally record an interrupt has occurred.'''
    if __debug__: log(f'interrupting wait')
    __waiter.set()


def interrupted():
    '''Return True if interrupt() has been called and not cleared.'''
    return __waiter.is_set()


def raise_for_interrupts():
    '''Check whether an interrupt occurred; if so, raise an exception.'''
    if interrupted():
        if __debug__: log(f'raising {__exception}')
        raise __exception


def reset():
    '''Clear the internal marker that an interrupt occurred.'''
    if __debug__: log(f'clearing wait')
    __waiter.clear()



#!/usr/bin/env python3
# =============================================================================
# @file    setup.py
# @brief   Installation setup file
# @author  Michael Hucka <mhucka@caltech.edu>
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/commonpy
#
# Note: configuration metadata is maintained in setup.cfg.  This file exists
# primarily to hook in setup.cfg and requirements.txt.
#
# =============================================================================

import os
from   os import path
from   setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'requirements.txt')) as f:
    reqs = f.read().rstrip().splitlines()

setup(
    setup_requires = ['wheel'],
    install_requires = reqs,
)



'''
file_utils.py: utilities for working with files.

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2019-2020 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import inspect
import os
from os.path import exists, isdir, isfile, dirname, relpath, realpath, splitext
import shutil
import subprocess
import sys
import tempfile
import webbrowser

if __debug__:
    from sidetrack import log


# Main functions.
# .............................................................................

def readable(dest):
    '''Returns True if the given 'dest' is accessible and readable.'''
    return os.access(dest, os.F_OK | os.R_OK)


def writable(dest):
    '''Returns True if the destination is writable.'''

    # Helper function to test if a directory is writable.
    def dir_writable(dir):
        # This is based on the following Stack Overflow answer by user "zak":
        # https://stackoverflow.com/a/25868839/743730
        try:
            testfile = tempfile.TemporaryFile(dir = dir)
            testfile.close()
        except (OSError, IOError) as e:
            return False
        return True

    if exists(dest) and not isdir(dest):
        # Path is an existing file.
        return os.access(dest, os.F_OK | os.W_OK)
    elif isdir(dest):
        # Path itself is an existing directory.  Is it writable?
        return dir_writable(dest)
    else:
        # Path is a file but doesn't exist yet. Can we write to the parent dir?
        return dir_writable(dirname(dest))


def nonempty(dest):
    '''Returns True if the file is not empty.'''
    return os.stat(dest).st_size != 0


def files_in_directory(dir, extensions = None, recursive = True):
    if not isdir(dir):
        return []
    if not readable(dir):
        return []
    if __debug__: log(f'reading directory {dir}')
    files = []
    for item in os.listdir(dir):
        full_path = path.join(dir, item)
        if isfile(full_path) and readable(full_path):
            if not extensions or filename_extension(item) in extensions:
                files.append(full_path)
        elif isdir(full_path) and recursive:
            files += files_in_directory(full_path, extensions)
    return sorted(files)


def filename_basename(file):
    parts = file.rpartition('.')
    if len(parts) > 1:
        return ''.join(parts[:-1]).rstrip('.') if parts[0] else file
    else:
        return file


def filename_extension(file):
    parts = file.rpartition('.')
    if parts[0] != '' and parts[1] != '':
        return '.' + parts[-1].lower()
    else:
        return ''


def alt_extension(filepath, ext):
    '''Returns the 'filepath' with the extension replaced by 'ext'.  The
    extension given in 'ext' should NOT have a leading period: that is, it
    should be "foo", not ".foo".'''
    return splitext(filepath)[0] + '.' + ext


def filtered_by_extensions(item_list, endings):
    if not item_list:
        return []
    if not endings:
        return item_list
    results = item_list
    for ending in endings:
        results = list(filter(lambda name: ending not in name.lower(), results))
    return results


def relative(file):
    '''Returns a path that is relative to the current directory.  If the
    relative path would require more than one parent step (i.e., ../../*
    instead of ../*) then it will return an absolute path instead.  If the
    argument is actuall a file path, it will return it unchanged.'''
    if is_url(file):
        return file
    candidate = relpath(file, os.getcwd())
    if not candidate.startswith('../..'):
        return candidate
    else:
        return realpath(candidate)


def rename_existing(file):
    '''Renames 'file' to 'file.bak'.'''

    def rename(f):
        backup = f + '.bak'
        # If we fail, we just give up instead of throwing an exception.
        try:
            os.rename(f, backup)
            if __debug__: log(f'renamed {file} to {backup}')
        except:
            try:
                delete_existing(backup)
                os.rename(f, backup)
            except:
                if __debug__: log(f'failed to delete {backup}')
                if __debug__: log(f'failed to rename {file} to {backup}')

    if exists(file):
        rename(file)
        return
    full_path = path.join(os.getcwd(), file)
    if exists(full_path):
        rename(full_path)
        return


def delete_existing(file):
    '''Delete the given file.'''
    # Check if it's actually a directory.
    if isdir(file):
        if __debug__: log(f'doing rmtree on directory {file}')
        try:
            shutil.rmtree(file)
        except:
            if __debug__: log(f'unable to rmtree {file}; will try renaming')
            try:
                rename_existing(file)
            except:
                if __debug__: log(f'unable to rmtree or rename {file}')
    else:
        if __debug__: log(f'doing os.remove on file {file}')
        os.remove(file)


def file_in_use(file):
    '''Returns True if the given 'file' appears to be in use.  Note: this only
    works on Windows, currently.
    '''
    if not exists(file):
        return False
    if sys.platform.startswith('win'):
        # This is a hack, and it really only works for this purpose on Windows.
        try:
            os.rename(file, file)
            return False
        except:
            return True
    return False


def copy_file(src, dst):
    '''Copies a file from "src" to "dst".'''
    if __debug__: log(f'copying file {src} to {dst}')
    shutil.copy2(src, dst, follow_symlinks = True)


def open_file(file):
    '''Opens document with default application in Python.'''
    # Code originally from https://stackoverflow.com/a/435669/743730
    if __debug__: log(f'opening file {file}')
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', file))
    elif os.name == 'nt':
        os.startfile(file)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', file))


def open_url(url):
    '''Opens the given 'url' in a web browser using the current platform's
    default approach.'''
    if __debug__: log(f'opening url {url}')
    webbrowser.open(url)
